// wuiobjdb.h - Definitions for WUIman OBJect DataBase.
// Copyright (c) 1993 by Ron Burk

#ifndef WUIOBJDB_H
#define WUIOBJDB_H

#if !(defined(_INC_WINDOWS) || defined(__WINDOWS_H))
    #include <windows.h>
#endif
#include "wuistd.h"
#include "wuiname.h"
#include "wuiargs.h"
#include "wuilist.h"
#include "wuipath.h"

class   AWuiObject;

typedef AWuiObject *(*WUI_INITFUNC)(const TWuiName Name);
typedef void        (*WUI_FINIFUNC)(const TWuiName Name);

class   TWuiMaster;
class   TWuiRegister
    {
public:
    TWuiRegister(const TWuiName Name,
       WUI_INITFUNC Constructor, WUI_FINIFUNC Destructor=0);
   ~TWuiRegister();
private:
    void    RegisterMaster(AWuiObject *MasterCopy);
    TWuiMaster   *Master;
    };


// various TWuiName(s) useful for WUIMAN class implementation

class   TWuiDatabase;
class   AWuiObject;
class   AWuiAttribute;
class   AWuiProperty;
class   TWuiObject;
class   TWuiRegister;
class   TWuiIntPropertyBit;

//TWuiTrace - Class to support function tracing
class   TWuiTrace
    {
public:
    TWuiTrace();
   ~TWuiTrace();
private:
    static int TraceLevel;
    };

// TWuiObjects - a list of type TWuiObject
class   TWuiObjects : public TWuiGenericList
    {
public:
    TWuiObjects(int Size=0);
    AWuiObject *Insert(AWuiObject *NewObject, const TWuiName To);
    AWuiObject *Insert(AWuiObject *NewObject, int Position=-1);
    AWuiObject *Get(const TWuiName Which);
    AWuiObject *Get(int Position);
    AWuiObject *Remove(const TWuiName Which);
    AWuiObject *Remove(int Position = -1);
    int         NObjects();
private:
    int         Find(const TWuiName Which);
    };

inline
AWuiObject *TWuiObjects::Remove(int Position)
    {   MEMBERASSERT();
    return (AWuiObject *)TWuiGenericList::Remove(Position);
    }
inline
AWuiObject *TWuiObjects::Get(int Position)
    {   MEMBERASSERT();
    return (AWuiObject *)TWuiGenericList::Get(Position);
    }
inline
int     TWuiObjects::NObjects()
    {   MEMBERASSERT();
    return TWuiGenericList::NElements();
    }


class   AWuiObject
    {
public:
    friend class TWuiRegister;

    AWuiObject(const TWuiName ObjectName);
    virtual              ~AWuiObject();
    virtual AWuiObject   *Clone(const TWuiName NewName,
                           int Depth=0)=0;
    virtual void          Copy(const AWuiObject *Other,
                                 int Depth);
    virtual long          Load(TWuiPath Path, const char *Value);
    virtual TWuiObjects  *Children() const;
    AWuiAttribute        *GetAttribute(const TWuiName Name) const;
    AWuiProperty         *GetProperty(const TWuiName Name) const;
    TWuiObject           *GetChild(const TWuiName Name) const;
    virtual TWuiObject   *GetChild(int Index, const TWuiName
                                  Order=(char *)NULL) const;
    virtual TWuiObjects  *Attributes() const;
    virtual long          Get(TWuiPath Path, TWuiArgs *Args,
                            char *Buffer, size_t MaxLength);
    virtual long          Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer);
    long                  Broadcast(TWuiObjects *List,
                              TWuiPath Path, TWuiArgs *Args,
                                       const char *Buffer=0,
                             size_t MaxLength=0, int Set=1);
    void                  Trace(const char *Function,
                              TWuiPath Path, TWuiArgs *Args);

    TWuiName              Name() const;
    virtual long          Rename(const TWuiName NewName);
    virtual TWuiName      ClassName() const = 0;
    virtual TWuiName      ParentClassName() const = 0;
    virtual int           IsA(const TWuiName Class,
                             const TWuiName Method=(char *)0) const = 0;
    virtual long          Save(TWuiPath Path, int Depth) const;
    long    DefineMethod(const TWuiName Method);

    static  TWuiObject   *NewObject(const TWuiName Type,
                                       const TWuiName Name);
    static  AWuiAttribute*NewAttribute(const TWuiName Type,
                                       const TWuiName Name);
    static  AWuiObject   *Root;
    static  TWuiDatabase *Database;
    static  TWuiObjects  *MasterObjects;
    static  int           TraceFlag;



    static TWuiNameBase M_NASTY_POINTER;

    static TWuiNameBase M_OBJECTCLONE;
    static TWuiNameBase M_OBJECTSAVE;
    static TWuiNameBase M_OBJECTDUMP;
    static TWuiNameBase M_OBJECTDELETE;
    static TWuiNameBase M_OBJECTISA;
    static TWuiNameBase M_OBJECTRENAME;
    static TWuiNameBase M_OBJECTCLASS;
    static TWuiNameBase M_OBJECTPARENT;
    static TWuiNameBase M_OBJECTCHILD;
    static TWuiNameBase M_OBJECTATTRIBUTE;
    static TWuiNameBase M_EDITATTR;

    static TWuiNameBase P_CLASS;
    static TWuiNameBase P_PARENTCLASS;
    static TWuiNameBase P_COMPILER;
    static TWuiNameBase P_WUIMANVERSION;
    static TWuiNameBase P_OBJECTORDER;

    static TWuiNameBase P_PERSISTENT;
    static TWuiNameBase P_INHERITED;
    static TWuiNameBase P_INHERITABLE;
    static TWuiNameBase P_TRACE;
    static TWuiNameBase P_READONLY;
    static TWuiNameBase P_REFERENCE;

    static TWuiNameBase S_CLASSES;
    static TWuiNameBase S_ROOT;

    static TWuiNameBase S_INHERITED;
    static TWuiNameBase S_INHERITABLE;
    static TWuiNameBase S_TRACE;
    static TWuiNameBase S_PERSISTENT;
    static TWuiNameBase S_READONLY;

    static TWuiNameBase C_ROOT;
    static TWuiNameBase C_TROOT;
    static TWuiNameBase C_TCONTAINER;
    static TWuiNameBase C_TORDEREDCONTAINER;
    static TWuiNameBase C_TNONCONTAINER;
    static TWuiNameBase C_TOBJECT;
    static TWuiNameBase C_TATTRIBUTE;
    static TWuiNameBase C_TMETHOD;
    static TWuiNameBase C_TPROPERTY;
    static TWuiNameBase C_TSTRINGPROPERTY;
    static TWuiNameBase C_TINTPROPERTY;
    static TWuiNameBase C_TLONGPROPERTY;
    static TWuiNameBase C_TDOUBLEPROPERTY;
    static TWuiNameBase C_TINTERNALBIT;

    static TWuiIntPropertyBit *P_Persistent;
    static TWuiIntPropertyBit *P_Inherited;
    static TWuiIntPropertyBit *P_Inheritable;
    static TWuiIntPropertyBit *P_Trace;
    static TWuiIntPropertyBit *P_ReadOnly;

    static TWuiObjects *PropertyProperties;

    static AWuiObject *ClassConstructor(const TWuiName Name);
    static void        ClassDestructor(const TWuiName Name);
    // flags are for use only by core WuiObject classes
protected:
    enum FLAGS
        {
        REFERENCE   = 0x01,
        READONLY    = 0x02,     PERSISTENT  = 0x04,
        INHERITED   = 0x08,     INHERITABLE = 0x10,
        INDIRECT    = 0x20,     TRACE       = 0x40,
        NORMAL      = (PERSISTENT|INHERITABLE|INHERITED)
        };
    int         GetFlags(int Flags=-1) const;
    int         SetFlags(int Flags=-1, int On=1);
    int         ClearFlags(int Flags=-1);
    int         NameToFlag(const TWuiName Name) const;
    void        DumpObject();

private:
    long    Clone(TWuiPath Path, int Depth, char *ToPath, const TWuiName Name);
    static AWuiObject   *Clone(const TWuiName Type,
                                       const TWuiName Name);
    TWuiName              Name_;
    int                   Flags_;
    static int            Initialized_;
    static  TWuiObjects  *Methods;
    };

inline
TWuiName AWuiObject::Name() const
    {   MEMBERASSERT();
    return Name_;
    }


class AWuiModifiableAttribute;

// TWuiObject - base class for non-attribute objects
// Main goal is to handle object persistence
class   TWuiObject : public AWuiObject
    {
public:
    TWuiObject(const TWuiName ObjectName, int NAttributes=0);
    virtual AWuiObject   *Clone(const TWuiName NewName,
                           int Depth=0);
    virtual void          Copy(const AWuiObject *Other,
                                 int Depth);
    virtual long          Save(TWuiPath Path, int Depth) const;
    virtual TWuiName      ClassName() const;
    virtual TWuiName      ParentClassName() const;
    virtual TWuiObjects  *Attributes() const;
    virtual int           IsA(const TWuiName Class,
                        const TWuiName Method=(char *)NULL) const;
    virtual long          Set(TWuiPath Path, TWuiArgs *Args,
                                        const char *Buffer);
    long            DefineAttribute(AWuiAttribute *Attribute);
private:
    TWuiObjects          *Attributes_;
    };

class   TWuiNonContainer : public TWuiObject
    {
public:
    TWuiNonContainer(const TWuiName ObjectName,
                                int NAttributes=0);
    virtual TWuiName      ClassName() const;
    virtual TWuiName      ParentClassName() const;
    virtual int           IsA(const TWuiName Class,
                        const TWuiName Method=(char *)NULL) const;
    virtual AWuiObject   *Clone(const TWuiName NewName,
                           int Depth=0);
    virtual void          Copy(const AWuiObject *Other,
                                 int Depth);
    static AWuiObject *ClassConstructor(const TWuiName Name);
    };

#ifdef __BORLANDC__
// Borland 3.1 generates bad code for this
static TWuiRegister RegisterAWuiObject(AWuiObject::C_TROOT,
        AWuiObject::ClassConstructor, AWuiObject::ClassDestructor);
#else
static TWuiRegister RegisterAWuiObject("",
        AWuiObject::ClassConstructor, AWuiObject::ClassDestructor);
#endif

static TWuiRegister RegisterNonContainer(
                                AWuiObject::C_TNONCONTAINER,
                        TWuiNonContainer::ClassConstructor);

// TWuiContainer - An AWuiObject that supports children and attributes.
class   TWuiContainer : public TWuiObject
    {
public:
    TWuiContainer(const TWuiName ObjectName,
                        int NAttributes=0, int NChildren=0);
    virtual AWuiObject   *Clone(const TWuiName NewName,
                           int Depth=0);
    virtual void          Copy(const AWuiObject *Other,
                                 int Depth);
    virtual TWuiName      ClassName() const;
    virtual TWuiName      ParentClassName() const;
    virtual TWuiObjects  *Children() const;
    virtual int           IsA(const TWuiName Class,
                        const TWuiName Method=(char *)NULL) const;
    virtual long          Set(TWuiPath Path, TWuiArgs *Args,
                                        const char *Buffer);
    static AWuiObject *ClassConstructor(const TWuiName Name);
private:
    TWuiObjects          *Children_;
    };

static TWuiRegister RegisterContainer(
                                   AWuiObject::C_TCONTAINER,
                           TWuiContainer::ClassConstructor);

// TWuiOrderedContainer - A container that imposes order
class   TWuiOrderedContainer : public TWuiContainer
    {
public:
    TWuiOrderedContainer(const TWuiName ObjectName,
                        int NAttributes=0, int NChildren=0);
   ~TWuiOrderedContainer();
    virtual AWuiObject   *Clone(const TWuiName NewName,
                                               int Depth=0);
    virtual void          Copy(const AWuiObject *Other,
                                 int Depth);
    virtual long          Load(TWuiPath Path, const char *Value);
    virtual TWuiName      ClassName() const;
    virtual TWuiName      ParentClassName() const;
    virtual int           IsA(const TWuiName Class,
                        const TWuiName Method=(char *)NULL) const;
    virtual long          Set(TWuiPath Path, TWuiArgs *Args,
                                        const char *Buffer);
    virtual TWuiObject   *GetChild(int Index, const TWuiName
                                  Order=(char *)NULL) const;
    static AWuiObject*ClassConstructor(const TWuiName Name);
    static const char SEPARATOR;
private:
    void        NewObjectOrder(TWuiPath Path);
    TWuiName   *ParseOrderString(const char *String, int *N);
    void        SetMruList(const TWuiName Property);
    int         IsOrdered(const TWuiName Property);
    void        RenumberMru(TWuiPath Path);

    char       *ObjectOrder;
    int         NOrdered;
    TWuiName   *PropertyNames;
    TWuiName    MruProperty;
    TWuiObjects*MruList;
    };

static TWuiRegister RegisterOrderedContainer(
                                   AWuiObject::C_TORDEREDCONTAINER,
                           TWuiOrderedContainer::ClassConstructor);



class   AWuiAttribute : public AWuiObject
    {
public:
    AWuiAttribute(const TWuiName ObjectName);
    virtual TWuiName ClassName() const;
    virtual TWuiName ParentClassName() const;
    virtual int      IsA(const TWuiName Class,
                             const TWuiName Method=(char *)0) const;
protected:
    void             EnforceName(char Prefix);
    };

class   AWuiModifiableAttribute : public AWuiAttribute
    {
public:
    AWuiModifiableAttribute(const TWuiName ObjectName,
                                        int Flags=NORMAL);
    virtual long          Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer);
    virtual long          Get(TWuiPath Path, TWuiArgs *Args,
                            char *Buffer, size_t MaxLength);
    virtual TWuiObjects  *Attributes() const;
//    virtual long          Load(TWuiPath Path, const char *Value);
    virtual long          Save(TWuiPath Path, int Depth) const;

    virtual void  BindAddress(void *Address);

protected:
    virtual long AttributeValue(char **Buffer) const = 0;
    };

class   AWuiProperty : public AWuiModifiableAttribute
    {
public:
    AWuiProperty(const TWuiName ObjectName,
                                        int Flags=NORMAL);
    virtual int      IsA(const TWuiName Class,
                     const TWuiName Method=(char *)0) const;

    virtual long    GetLong() const;
    virtual int     GetInt() const;
    virtual size_t  GetString(char *To, size_t MaxLength) const;
    };

class   TWuiEvent : public AWuiModifiableAttribute
    {
public:
    TWuiEvent(const TWuiName ObjectName,
                                        int Flags=NORMAL);
private:
    virtual long AttributeValue(char **Buffer) const;
    };

class   TWuiMethod : public AWuiAttribute
    {
public:
    TWuiMethod(const TWuiName ObjectName);
    virtual AWuiObject   *Clone(const TWuiName NewName,
                           int Depth=0);
    virtual void          Copy(const AWuiObject *Other,
                                 int Depth);
    virtual TWuiName    ClassName() const;
    virtual TWuiName    ParentClassName() const;
protected:
    virtual long AttributeValue(char **Buffer) const;
    };

class TWuiIntProperty : public AWuiProperty
    {
public:
    TWuiIntProperty(const TWuiName Name, int *Address=0);
    virtual AWuiObject   *Clone(const TWuiName NewName,
                           int Depth=0);
    virtual void          Copy(const AWuiObject *Other,
                                 int Depth);
    virtual TWuiName    ClassName() const;
    virtual TWuiName    ParentClassName() const;
    virtual int           IsA(const TWuiName Class,
                             const TWuiName Method=(char *)0) const;
    virtual long        Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer);
    static  AWuiObject *ClassConstructor(const TWuiName Name);
    virtual long    Load(TWuiPath Path, const char *Value);
    virtual long    GetLong() const;
    virtual int     GetInt() const;
protected:
    virtual long AttributeValue(char **Buffer) const;
private:
    virtual void  BindAddress(void *Address);
    long          SetValue(const char *Value);
    int          *GetAddress() const;
    union
        {
        int     IntValue;
        int    *IntAddress;
        };
    };

static TWuiRegister RegisterIntProperty(
                                AWuiObject::C_TINTPROPERTY,
                         TWuiIntProperty::ClassConstructor);

class TWuiLongProperty : public AWuiProperty
    {
public:
    TWuiLongProperty(const TWuiName Name, long *Address=0);
    virtual AWuiObject   *Clone(const TWuiName NewName,
                           int Depth=0);
    virtual void          Copy(const AWuiObject *Other,
                                 int Depth);
    virtual TWuiName    ClassName() const;
    virtual TWuiName    ParentClassName() const;
    virtual int           IsA(const TWuiName Class,
                             const TWuiName Method=(char *)0) const;
    virtual long        Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer);
    virtual long    Load(TWuiPath Path, const char *Value);
    virtual long    GetLong() const;
    virtual int     GetInt() const;
    static  AWuiObject *ClassConstructor(const TWuiName Name);
private:
    virtual long AttributeValue(char **Buffer) const;
    virtual void  BindAddress(void *Address);
    long          SetValue(const char *Value);
    long         *GetAddress() const;
    union
        {
        long    LongValue;
        long   *LongAddress;
        };
    };
static TWuiRegister RegisterLongProperty(
                                AWuiObject::C_TLONGPROPERTY,
                         TWuiLongProperty::ClassConstructor);

#ifdef  WUI_FLOAT

class TWuiDoubleProperty : public AWuiProperty
    {
public:
    TWuiDoubleProperty(const TWuiName Name, double *Address=0);
    virtual AWuiObject   *Clone(const TWuiName NewName,
                           int Depth=0);
    virtual void          Copy(const AWuiObject *Other,
                                 int Depth);
    virtual TWuiName    ClassName() const;
    virtual TWuiName    ParentClassName() const;
    virtual int           IsA(const TWuiName Class,
                             const TWuiName Method=(char *)0) const;
    virtual long        Load(TWuiPath Path, const char *Value);
    virtual long        Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer);
    static  AWuiObject *ClassConstructor(const TWuiName Name);
protected:
    virtual long AttributeValue(char **Buffer) const;
private:
    long    SetValue(const char *Value);
    virtual void  BindAddress(void *Address);
    union
        {
        double  DoubleValue;
        double *DoubleAddress;
        };
    };
static TWuiRegister RegisterDoubleProperty(
                                AWuiObject::C_TDOUBLEPROPERTY,
                         TWuiDoubleProperty::ClassConstructor);

#endif  // WUI_FLOAT

class TWuiStringProperty : public AWuiProperty
    {
public:
    TWuiStringProperty(const TWuiName Name, char **Address=0);
   ~TWuiStringProperty();
    virtual long    Load(TWuiPath Path, const char *Value);
    virtual long          Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer);
    virtual AWuiObject   *Clone(const TWuiName NewName,
                           int Depth=0);
    virtual void          Copy(const AWuiObject *Other,
                                 int Depth);
    virtual TWuiName    ClassName() const;
    virtual TWuiName    ParentClassName() const;
    virtual int         IsA(const TWuiName Class,
                             const TWuiName Method=(char *)0) const;
    virtual void        BindAddress(void *Address);
    static  AWuiObject *ClassConstructor(const TWuiName Name);
protected:
    virtual long AttributeValue(char **Buffer) const;
private:
    long        SetValue(const char *Value);
    union
        {
        char   *StringValue;
        char  **StringAddress;
        };
    };

static TWuiRegister RegisterStringProperty(
                                AWuiObject::C_TSTRINGPROPERTY,
                         TWuiStringProperty::ClassConstructor);


class TWuiIntPropertyBit : public AWuiProperty
    {
public:
    TWuiIntPropertyBit(const TWuiName Name);
    virtual AWuiObject   *Clone(const TWuiName NewName,
                           int Depth=0);
    virtual void          Copy(const AWuiObject *Other,
                                 int Depth);
    virtual TWuiName    ClassName() const;
    virtual TWuiName    ParentClassName() const;
    virtual long	Set(TWuiPath Path, TWuiArgs *Args,
                                         const char *Buffer);
protected:
    virtual long AttributeValue(char **Buffer) const;
private:
    };




class   TWuiRoot : public TWuiContainer
   {
public:
    TWuiRoot(const TWuiName ObjectName);
    virtual TWuiName    ClassName() const;
    virtual TWuiName    ParentClassName() const;
    static  AWuiObject *ClassConstructor(const TWuiName Name);
private:
    char               *Compiler;
    char               *WuimanVersion;
    };


#endif
